
import random as r

# an alignment like
# TCA--CCCAT
# GGAGGTTGGT
# is represented in a list as ['TCA--CCCAT','GGAGGTTGGT']



def make_rnd_DNA_seq(len):
    """returns a random sequence of nucleotides, whose length is 'len'"""
    bases='ATCG'
    seq = ''
    while len > 0:
        c = r.choice(bases)
        seq += c
        len -= 1
    return seq



# a similarity matrix is represented by a dict of dicts, so that the sililarity
# of base T in seq 1 and G in seq 2 can be referred as, e.g., sim['T']['G']

# some possible similarity matrices are reported hereafter

sim_m_0 = {'A': {'A':10, 'G':-1, 'C':-3, 'T':-4}, \
    'G': {'A':-1, 'G':7, 'C':-5, 'T':-3}, \
    'C': {'A':-3, 'G':-5, 'C':9, 'T':0}, \
    'T': {'A':-4, 'G':-3, 'C':0, 'T':8}}

sim_m_1 = {'A': {'A':10, 'G':-6, 'C':-6, 'T':-6}, \
    'G': {'A':-6, 'G':10, 'C':-6, 'T':-6}, \
    'C': {'A':-6, 'G':-6, 'C':10, 'T':-6}, \
    'T': {'A':-6, 'G':-6, 'C':-6, 'T':10}}

#the next is the one used in BLAST
sim_BLAST = {'A': {'A':5, 'G':-4, 'C':-4, 'T':-4}, \
    'G': {'A':-4, 'G':5, 'C':-4, 'T':-4}, \
    'C': {'A':-4, 'G':-4, 'C':5, 'T':-4}, \
    'T': {'A':-4, 'G':-4, 'C':-4, 'T':5}}

#the next one considers transitions/transvections
sim_TR_TR = {'A': {'A':1, 'G':-1, 'C':-5, 'T':-5}, \
    'G': {'A':-1, 'G':1, 'C':-5, 'T':-5}, \
    'C': {'A':-5, 'G':-5, 'C':1, 'T':-1}, \
    'T': {'A':-5, 'G':-5, 'C':-1, 'T':1}}

#the next one corresponds to -1*(Levenshtein) (-edit_distance)
sim_neg_Leven = {'A': {'A':0, 'G':-1, 'C':-1, 'T':-1}, \
    'G': {'A':-1, 'G':0, 'C':-1, 'T':-1}, \
    'C': {'A':-1, 'G':-1, 'C':0, 'T':-1}, \
    'T': {'A':-1, 'G':-1, 'C':-1, 'T':0}}


def score_alignment(al, simil_matrix, gp):
    """The score of alignment 'al', given simil. matrix & gap penalty"""
    al_score = 0
    for i in xrange(len(al[0])):
        if al[0][i] != '-' and al[1][i] != '-':
            al_score += simil_matrix[al[0][i]][al[1][i]]
        else: al_score += gp
    return al_score

def print_alignment(al):
    """
    A nice, standard textual output for an alignment
    """
    string_upper=al[0]
    string_lower=al[1]
    string_in_between=''
    for sx,sy in zip(al[0],al[1]):
        if sx == sy:
            string_in_between+="|"
        else:
            string_in_between+=" "
    print "  ",string_upper
    print "  ",string_in_between
    print "  ",string_lower



def compute_matrix_F(a, b, similarity_matrix, gap_penalty):
    """
    Returns the score matrix F for two sequences a and b to be aligned.
    F is the matrix for partial scores.
    The matrix is represented as a list of lists, with #columns =len(a)+1
    and #rows = len(b)+1
    """

    len_a = len(a)
    len_b = len(b)

    matrix_F = []

    for i in xrange(len_b + 1):
        temp = []
        for j in xrange(len_a + 1):
            if i == 0:
                temp.append(gap_penalty * j)
            elif j == 0:
                temp.append(gap_penalty * i)
            else:
                temp.append(0)

        matrix_F.append(temp)

    #computation of scores

    for i in xrange(1, len_b + 1):
        for j in xrange(1, len_a + 1):
            val1 = matrix_F[i-1][j-1] + similarity_matrix[a[j-1]][b[i-1]]
            val2 = matrix_F[i][j-1] + gap_penalty
            val3 = matrix_F[i-1][j] + gap_penalty
            val = max(val1, val2, val3)
            matrix_F[i][j] = val

    return matrix_F




def g_align_nw_simple(a, b, simil_matrix, gap_penalty):

    len_a = len(a)
    len_b = len(b)

    matrix_F = compute_matrix_F(a,b,simil_matrix, gap_penalty)

    al_upper_seq = ''
    al_lower_seq = ''
    i = len(b)
    j = len(a)
    while i > 0 and j > 0:
        score = matrix_F[i][j]
        score_diag = matrix_F[i - 1][j - 1]
        score_up = matrix_F[i-1][j]
        score_left = matrix_F[i][j-1]
        if score == score_diag + simil_matrix[a[j-1]][b[i-1]]:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1
            j -= 1
        elif score == score_left + gap_penalty:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = '-' + al_lower_seq
            j -= 1
        elif score == score_up + gap_penalty:
            al_upper_seq = '-' + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1

    #the following is done to go to [0,0] once the first row/column is reached
    while j > 0:
        al_upper_seq = a[j-1] + al_upper_seq
        al_lower_seq = '-' + al_lower_seq
        j -= 1
    while i > 0:
        al_upper_seq = '-' + al_upper_seq
        al_lower_seq = b[i-1] + al_lower_seq
        i -= 1

    return [al_upper_seq, al_lower_seq]



def g_align_nw(a, b, simil_matrix, gap_penalty):
    """
    global alignment with Needleman Wunsch;
    returns a list with all the alignments with max score
    """

    matrix_F = compute_matrix_F(a, b, simil_matrix, gap_penalty)

    all_alignments = []

    def find_alignments(all_alignments, matrix_F, index_a, index_b, partial_a, partial_b):
        
        i = index_b
        j = index_a

        if(i > 0 and j > 0):
            score = matrix_F[i][j]
            
            score_diag = matrix_F[i - 1][j - 1]
            score_up = matrix_F[i-1][j]
            score_left = matrix_F[i][j-1]

            if score == score_diag + simil_matrix[a[j-1]][b[i-1]]:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i-1, new_partial_a[:], new_partial_b[:])

            if score == score_left + gap_penalty:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = '-' + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i, new_partial_a[:], new_partial_b[:])

            if score == score_up + gap_penalty:
                new_partial_a = '-' + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j, i-1, new_partial_a[:], new_partial_b[:])
        else:

            while j > 0:
                partial_a = a[j-1] + partial_a
                partial_b = '-' + partial_b
                j -= 1
            while i > 0:
                partial_a = '-' + partial_a
                partial_b = b[i-1] + partial_b
                i -= 1

            current_alignment = []
            current_alignment.append(partial_a)
            current_alignment.append(partial_b)

            all_alignments.append(current_alignment)

        return all_alignments


    return find_alignments(all_alignments, matrix_F, len(matrix_F[0])-1, len(matrix_F)-1, '', '')



#######################################################
# the following is dedicated to semiglobal alignments #
#######################################################


def compute_matrix_F_sga(a, b, similarity_matrix, gap_penalty):
    """
    The same as matrix F, but for semilocal alignments
    """

    len_a = len(a)
    len_b = len(b)

    matrix_F = []

    for i in xrange(len_b + 1):
        temp = []
        for j in xrange(len_a + 1):
            temp.append(0)

        matrix_F.append(temp)

    #computation of scores

    for i in xrange(1, len_b + 1):
        for j in xrange(1, len_a + 1):
            penalty=gap_penalty
            if i ==len_b or j==len_a:
                penalty=0
            val1 = matrix_F[i-1][j-1] + similarity_matrix[a[j-1]][b[i-1]]
            val2 = matrix_F[i][j-1]+penalty
            val3 = matrix_F[i-1][j]+penalty
            
            val = max(val1, val2, val3)
            matrix_F[i][j] = val

    return matrix_F



def g_align_sga_simple(a, b, simil_matrix, gap_penalty):
    """
    The same as the simple Needleman Wunsch, but for semi-global alignment
    """

    len_a = len(a)
    len_b = len(b)

    matrix_F = compute_matrix_F_sga(a,b,simil_matrix, gap_penalty)

    al_upper_seq = ''
    al_lower_seq = ''
    i = len_b
    j = len_a
    while i > 0 and j > 0:
        score = matrix_F[i][j]
        score_diag = matrix_F[i - 1][j - 1]
        score_up = matrix_F[i-1][j]
        score_left = matrix_F[i][j-1]
        penalty=gap_penalty
        if i==len_b or j == len_a:
            penalty=0
            
        if score == score_diag + simil_matrix[a[j-1]][b[i-1]]:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1
            j -= 1
        elif score == score_left + penalty:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = '-' + al_lower_seq
            j -= 1
        elif score == score_up + penalty:
            al_upper_seq = '-' + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1

    #the following is done to go to [0,0] once the first row/column is reached
    while j > 0:
        al_upper_seq = a[j-1] + al_upper_seq
        al_lower_seq = '-' + al_lower_seq
        j -= 1
    while i > 0:
        al_upper_seq = '-' + al_upper_seq
        al_lower_seq = b[i-1] + al_lower_seq
        i -= 1

    return [al_upper_seq, al_lower_seq]





def g_align_sga(a, b, simil_matrix, gap_penalty):
    """
    The same as Needleman Wunsch, but for semi-global alignment
    """

    matrix_F = compute_matrix_F_sga(a, b, simil_matrix, gap_penalty)
    len_a=len(a)
    len_b=len(b)

    all_alignments = []

    def find_alignments(all_alignments, matrix_F, index_a, index_b, partial_a, partial_b):
        
        i = index_b
        j = index_a

        if(i > 0 and j > 0):
            score = matrix_F[i][j]
            
            score_diag = matrix_F[i - 1][j - 1]
            score_up = matrix_F[i-1][j]
            score_left = matrix_F[i][j-1]

            penalty=gap_penalty
            if i==len_b or j == len_a:
                penalty=0

            if score == score_diag + simil_matrix[a[j-1]][b[i-1]]:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i-1, new_partial_a[:], new_partial_b[:])

            if score == score_left + penalty:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = '-' + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i, new_partial_a[:], new_partial_b[:])

            if score == score_up + penalty:
                new_partial_a = '-' + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j, i-1, new_partial_a[:], new_partial_b[:])
        else:

            while j > 0:
                partial_a = a[j-1] + partial_a
                partial_b = '-' + partial_b
                j -= 1
            while i > 0:
                partial_a = '-' + partial_a
                partial_b = b[i-1] + partial_b
                i -= 1

            current_alignment = []
            current_alignment.append(partial_a)
            current_alignment.append(partial_b)

            all_alignments.append(current_alignment)

        return all_alignments


    return find_alignments(all_alignments, matrix_F, len(matrix_F[0])-1, len(matrix_F)-1, '', '')

